//-----------------------------------------------------------------------------
// MK3_2560_Servo_Mess
// Version 1.0
// 05.02.2013 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer das myAVR MK3 Board
// mit ATMega-2560 @ 16MHz
// Treiber zur Messung eines Modellbau-Servo-Signals
// Typische Impulszeiten = 1ms bis 2ms (Neutral = 1,5ms)
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_MK3_2560/"
// liegen und mit #include <UB_MK3_2560/MK3_2560_Servo_Mess.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : @MK3 (ATMega2560)
//            Timer-1 (16bit)
//            Port-D Bit-4 (als Trigger Input [ICP1])
// Belegt : @MK2 (ATMega8)
//            Timer-1 (16bit)
//            Port-B Bit-0 (als Trigger Input [ICP1])
//
// Hinweis :
//
// > die DEFINE-Schalter "SERVOMESS_HW_MODE" und "SERVOMESS_QUARZ_TYP"
//   muessen zur Hardware passen !!

//-----------------------------------------------------------------------------

#include <avr/interrupt.h> 


//-----------------------------------------------------------------------------
// Hardware auswahl
//-----------------------------------------------------------------------------
#define SERVOMESS_HW_MODE     0        // 0=MK3-Board, 1=MK1/MK2-Board/mySmartControl
#define SERVOMESS_QUARZ_TYP   0        // 0=16MHz, 1=8MHz, 2=3.6864MHz


//-----------------------------------------------------------------------------
// Globale Defines
//-----------------------------------------------------------------------------

#define SERVOMESS_IMPULS_OK     3000   // wenn Impulswert < 3000us dann Messung OK
#define SERVOMESS_WAITING      60000   // es wurde noch kein Impuls gemessen
#define SERVOMESS_IDLE         61000   // die Impulsmessung laeuft gerade
#define SERVOMESS_TOOHIGH      62000   // der Messwert ist > als SERVOMESS_IMPULS_OK
#define SERVOMESS_OVF          63000   // es gab einen Ueberlauf -> Messung neu gestartet


//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void MK3_INIT_SERVOMESS(void);             // Initialisierung
unsigned int MK3_SERVOMESS_READ_US(void);  // Gemessener Impulswert in us
//-----------------------------------------------------------------------------



//-----------------------------------------------------------------------------
// Port-Pin definition
//-----------------------------------------------------------------------------
#if SERVOMESS_HW_MODE==0
  // Hardware = myAVR MK3-Board
  #define MK3_DDR_ICP  DDRD    // DDR-Register vom ICP-Pin
  #define MK3_PORT_ICP PORTD   // Port vom ICP-Pin
  #define MK3_ICP_PIN  0x10    // Bit vom ICP-Pin
#else
  // Hardware = myAVR MK1/MK2-Board + mySmartControl
  #define MK3_DDR_ICP  DDRB    // DDR-Register vom ICP-Pin
  #define MK3_PORT_ICP PORTB   // Port vom ICP-Pin
  #define MK3_ICP_PIN  0x01    // Bit vom ICP-Pin
#endif

//-----------------------------------------------------------------------------
// interne defines
//-----------------------------------------------------------------------------
#if SERVOMESS_QUARZ_TYP==0
  // 16MHZ 
  #define SERVOMESS_FKT 2.0 
#elif SERVOMESS_QUARZ_TYP==1
  // 8MHz  
  #define SERVOMESS_FKT 1.0
#elif SERVOMESS_QUARZ_TYP==2
  // 3.6 MHz
  #define SERVOMESS_FKT 0.4608
#endif

//-----------------------------------------------------------------------------
// interne Enumerationen
//-----------------------------------------------------------------------------
typedef enum {
  MESS_WAIT =0,  // warte auf Signal
  MESS_RUN,      // Signal erkannt, warte auf Ende
  MESS_OK,       // Ende erkannt, Messung fertig
  MESS_OVF       // Signal erkannt aber Timer ueberlauf
}MESS_STATUS;


typedef enum {
  TRG_FLANKE_LO =0, // ICP triggert auf Lo-Flanke
  TRG_FLANKE_HI     // ICP triggert auf Hi-Flanke
}TRG_FLANKE;


//-----------------------------------------------------------------------------
// interne Globale Variabeln
//-----------------------------------------------------------------------------
typedef struct {
  TRG_FLANKE trigger;
  MESS_STATUS status;
  unsigned int messwert;
}ServoMess_Intern_t;
ServoMess_Intern_t ServoMess;


//-----------------------------------------------------------------------------
void MK3_INIT_SERVOMESS(void) {  // zum initialisieren (startet die Messung)

  ServoMess.trigger=TRG_FLANKE_HI;
  ServoMess.status=MESS_WAIT;
  ServoMess.messwert=0;

  // benutzter Timer : Timer-1 (16bit)
  // im Mode0 -> normal

  #if SERVOMESS_HW_MODE==0
    // Hardware = myAVR MK3-Board 
    TCCR1A=0x00;
    TCCR1B=0x00; // Mode 0
    TCCR1C=0x00;
  #else
    // Hardware = myAVR MK1/MK2-Board + mySmartControl
    TCCR1A=0x00;
    TCCR1B=0x00; // Mode 0
  #endif


  // ICP-Pin auf Eingang und Pull-Up einschalten
  MK3_DDR_ICP&=~MK3_ICP_PIN;
  MK3_PORT_ICP|=MK3_ICP_PIN;
  TCCR1B|=0x80;  // noise canceler enable
  // Messung starten
  TCCR1B|=0x40;  // trigger auf Hi-Flanke
  // ICP und OVF Interrupt aktivieren
  #if SERVOMESS_HW_MODE==0
    // Hardware = myAVR MK3-Board       
    TIMSK1|=0x24;  // ICP+OVF interrupt enable
  #else
    // Hardware = myAVR MK1/MK2-Board + mySmartControl
    TIMSK|=0x24;  // ICP+OVF interrupt enable
  #endif

  // Counter Werte init
  TCNT1H=0x00;
  TCNT1L=0x00;

  SREG|=0x80; // Alle Interrupts einschalten
}


//-----------------------------------------------------------------------------
// auslesen vom Servo-Impulswert (in us)
// wert liegt bei normalen Servos im bereich von 1000us bis 2000us
//
// wenn wert > SERVOMESS_IMPULS_OK -> Fehler
// moegliche Fehlerwerte :
//    SERVOMESS_WAITING : es wurde noch kein Impuls gemessen
//    SERVOMESS_IDLE    : die Impulsmessung laeuft gerade 
//    SERVOMESS_TOOHIGH : der Messwert ist > als SERVOMESS_IMPULS_OK
//    SERVOMESS_OVF     : es gab einen Ueberlauf -> Messung neu gestartet
//-----------------------------------------------------------------------------
unsigned int MK3_SERVOMESS_READ_US(void) {
  unsigned int ret_wert=0;

  if(ServoMess.status==MESS_WAIT) {
    ret_wert=SERVOMESS_WAITING;
  }
  else if(ServoMess.status==MESS_RUN) {
    ret_wert=SERVOMESS_IDLE;
  }
  else if(ServoMess.status==MESS_OK) {
    // Messwert in us wandeln
    ret_wert=ServoMess.messwert/SERVOMESS_FKT;
    if(ret_wert>SERVOMESS_IMPULS_OK) {
      ret_wert=SERVOMESS_TOOHIGH;
    }
    // naechste Messung starten
    ServoMess.status=MESS_WAIT;
    ServoMess.trigger=TRG_FLANKE_HI;
    TCCR1B|=0x40;  // trigger auf Hi-Flanke
  }
  else {
    ret_wert=SERVOMESS_OVF;
  }  
  
  return(ret_wert);      
}



//-----------------------------------------------------------------------------
// ISR von Timer-1 ICP
//-----------------------------------------------------------------------------
ISR (TIMER1_CAPT_vect) {                          // ISR : Timer1 ICP
  // bei Timer1 ICP event
  unsigned int uint;  

  // Messwert auslesen
  uint=ICR1;

  // Counter Werte zuruecksetzen
  TCNT1H=0x00;
  TCNT1L=0x00;

  // Trigger neu setzen
  if(ServoMess.trigger==TRG_FLANKE_HI) {    
    ServoMess.trigger=TRG_FLANKE_LO;
    TCCR1B&=~0x40;  // trigger auf Lo-Flanke    	
    // Counter starten mit Vorteiler
    TCCR1B|=0x02;	// Vorteiler = 8	
    ServoMess.status=MESS_RUN;
  }
  else {           
    if(ServoMess.status!=MESS_OK) {	
      ServoMess.messwert=uint;
      // counter stoppen
      TCCR1B&=~0x07;
      ServoMess.status=MESS_OK; 		
    }
  }   
}

//-----------------------------------------------------------------------------
// ISR von Timer-1 OVF
//-----------------------------------------------------------------------------
ISR (TIMER1_OVF_vect) {                          // ISR : Timer1 OVF
  // Timerueberlauf (keine Lo-Flanke gemessen)
  // counter stoppen
  TCCR1B&=~0x07;
  ServoMess.status=MESS_OVF;
  // naechste Messung starten  
  ServoMess.trigger=TRG_FLANKE_HI;
  TCCR1B|=0x40;  // trigger auf Hi-Flanke  
}



